from .base import (
    MappingPattern,
    RegexPattern,
    SequencePattern,
    SwitchPattern,
    UnionPattern,
)
from .config import lang
from .context import (
    Patterns,
    all_patterns,
    create_local_patterns,
    local_patterns,
    switch_local_patterns,
    reset_local_patterns
)
from .core import BasePattern, MatchMode, ValidateResult, set_unit
from .exception import MatchFailed
from .main import (
    DATETIME,
    EMAIL,
    FLOAT,
    HEX,
    HEX_COLOR,
    INTEGER,
    IP,
    NUMBER,
    URL,
    AnyOne,
    AnyString,
    Bind,
    type_parser,
)
from .util import AllParam, Empty, TPattern, generic_isinstance
